namespace Syntriax.Engine.Core;

/// <summary>
/// Represents an RGB color.
/// </summary>
/// <param name="r">Red value of the <see cref="ColorRGB"/>.</param>
/// <param name="g">Green value of the <see cref="ColorRGB"/>.</param>
/// <param name="b">Blue value of the <see cref="ColorRGB"/>.</param>
/// <remarks>
/// Initializes a new instance of the <see cref="ColorRGB"/> struct with the specified values.
/// </remarks>
[System.Diagnostics.DebuggerDisplay("{ToString(),nq}")]
public readonly struct ColorRGB(byte r, byte g, byte b)
{
    /// <summary>
    /// The Red value of the <see cref="ColorRGB"/>.
    /// </summary>
    public readonly byte R = r;

    /// <summary>
    /// The Green value of the <see cref="ColorRGB"/>.
    /// </summary>
    public readonly byte G = g;

    /// <summary>
    /// The Blue value of the <see cref="ColorRGB"/>.
    /// </summary>
    public readonly byte B = b;

    public static ColorRGB operator -(ColorRGB color) => new((byte)(255 - color.R), (byte)(255 - color.G), (byte)(255 - color.B));
    public static ColorRGB operator +(ColorRGB left, ColorRGB right) => new((byte)(left.R + right.R).Clamp(0, 255), (byte)(left.G + right.G).Clamp(0, 255), (byte)(left.B + right.B).Clamp(0, 255));
    public static ColorRGB operator -(ColorRGB left, ColorRGB right) => new((byte)(left.R - right.R).Clamp(0, 255), (byte)(left.G - right.G).Clamp(0, 255), (byte)(left.B - right.B).Clamp(0, 255));
    public static ColorRGB operator *(ColorRGB left, ColorRGB right) => new((byte)(left.R * right.R).Clamp(0, 255), (byte)(left.G * right.G).Clamp(0, 255), (byte)(left.B * right.B).Clamp(0, 255));
    public static ColorRGB operator *(ColorRGB color, float value) => new((byte)(color.R * value).Clamp(0, 255), (byte)(color.G * value).Clamp(0, 255), (byte)(color.B * value).Clamp(0, 255));
    public static ColorRGB operator *(float value, ColorRGB color) => new((byte)(color.R * value).Clamp(0, 255), (byte)(color.G * value).Clamp(0, 255), (byte)(color.B * value).Clamp(0, 255));
    public static ColorRGB operator /(ColorRGB color, float value) => new((byte)(color.R / value).Clamp(0, 255), (byte)(color.G / value).Clamp(0, 255), (byte)(color.B / value).Clamp(0, 255));
    public static bool operator ==(ColorRGB left, ColorRGB right) => left.R == right.R && left.G == right.G && left.B == right.B;
    public static bool operator !=(ColorRGB left, ColorRGB right) => left.R != right.R || left.G != right.G || left.B != right.B;

    public static implicit operator ColorRGB(ColorRGBA rgba) => new(rgba.R, rgba.G, rgba.B);
    public static implicit operator ColorRGB(ColorHSV hsv)
    {
        float hue = hsv.Hue * 360f;
        float chroma = hsv.Value * hsv.Saturation;
        float x = chroma * (1f - Math.Abs(hue / 60f % 2f - 1f));
        float m = hsv.Value - chroma;

        float r1 = 0f;
        float g1 = 0f;
        float b1 = 0f;

        if (hue < 60) { r1 = chroma; g1 = x; b1 = 0; }
        else if (hue < 120) { r1 = x; g1 = chroma; b1 = 0; }
        else if (hue < 180) { r1 = 0; g1 = chroma; b1 = x; }
        else if (hue < 240) { r1 = 0; g1 = x; b1 = chroma; }
        else if (hue < 300) { r1 = x; g1 = 0; b1 = chroma; }
        else if (hue <= 360) { r1 = chroma; g1 = 0; b1 = x; }

        byte r = (byte)Math.RoundToInt((r1 + m) * 255);
        byte g = (byte)Math.RoundToInt((g1 + m) * 255);
        byte b = (byte)Math.RoundToInt((b1 + m) * 255);

        return new(r, g, b);
    }

    /// <summary>
    /// Inverts the given <see cref="ColorRGB"/>.
    /// </summary>
    /// <param name="color">The <see cref="ColorRGB"/>.</param>
    /// <returns>The inverted <see cref="ColorRGB"/>.</returns>
    public static ColorRGB Invert(ColorRGB color) => -color;

    /// <summary>
    /// Adds two <see cref="ColorRGB"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="ColorRGB"/>.</param>
    /// <param name="right">The second <see cref="ColorRGB"/>.</param>
    /// <returns>The sum of the two <see cref="ColorRGB"/>s.</returns>
    public static ColorRGB Add(ColorRGB left, ColorRGB right) => left + right;

    /// <summary>
    /// Subtracts one <see cref="ColorRGB"/> from another.
    /// </summary>
    /// <param name="left">The <see cref="ColorRGB"/> to subtract from.</param>
    /// <param name="right">The <see cref="ColorRGB"/> to subtract.</param>
    /// <returns>The result of subtracting the second <see cref="ColorRGB"/> from the first.</returns>
    public static ColorRGB Subtract(ColorRGB left, ColorRGB right) => left - right;

    /// <summary>
    /// Multiplies a <see cref="ColorRGB"/> by a scalar value.
    /// </summary>
    /// <param name="color">The <see cref="ColorRGB"/>.</param>
    /// <param name="value">The scalar value.</param>
    /// <returns>The result of multiplying the <see cref="ColorRGB"/> by the scalar value.</returns>
    public static ColorRGB Multiply(ColorRGB color, float value) => color * value;

    /// <summary>
    /// Divides a <see cref="ColorRGB"/> by a scalar value.
    /// </summary>
    /// <param name="color">The <see cref="ColorRGB"/>.</param>
    /// <param name="value">The scalar value.</param>
    /// <returns>The result of dividing the <see cref="ColorRGB"/> by the scalar value.</returns>
    public static ColorRGB Divide(ColorRGB color, float value) => color / value;

    /// <summary>
    /// Calculates the <see cref="ColorRGB"/> from one point to another.
    /// </summary>
    /// <param name="from">The starting point.</param>
    /// <param name="to">The ending point.</param>
    /// <returns>The <see cref="ColorRGB"/> from the starting point to the ending point.</returns>
    public static ColorRGB FromTo(ColorRGB from, ColorRGB to) => to - from;

    /// <summary>
    /// Performs linear interpolation between two <see cref="ColorRGB"/>s.
    /// </summary>
    /// <param name="from">The starting <see cref="ColorRGB"/> (t = 0).</param>
    /// <param name="to">The ending <see cref="ColorRGB"/> (t = 1).</param>
    /// <param name="t">The interpolation parameter.</param>
    /// <returns>The interpolated <see cref="ColorRGB"/>.</returns>
    public static ColorRGB Lerp(ColorRGB from, ColorRGB to, float t) => from + FromTo(from, to) * t;

    /// <summary>
    /// Determines whether the specified object is equal to the current <see cref="ColorRGB"/>.
    /// </summary>
    /// <param name="obj">The object to compare with the current <see cref="ColorRGB"/>.</param>
    /// <returns><see cref="true"/> if the specified object is equal to the current <see cref="ColorRGB"/>; otherwise, <see cref="false"/>.</returns>
    public override bool Equals(object? obj) => obj is ColorRGB colorRGB && R.Equals(colorRGB.R) && G.Equals(colorRGB.G) && B.Equals(colorRGB.B);

    /// <summary>
    /// Generates a hash code for the <see cref="ColorRGB"/>.
    /// </summary>
    /// <returns>A hash code for the <see cref="ColorRGB"/>.</returns>
    public override int GetHashCode() => System.HashCode.Combine(R, G, B);

    /// <summary>
    /// Converts the <see cref="ColorRGB"/> to its string representation.
    /// </summary>
    /// <returns>A string representation of the <see cref="ColorRGB"/>.</returns>
    public override string ToString() => $"{nameof(ColorRGB)}({R}, {G}, {B})";
}

/// <summary>
/// Provides extension methods for <see cref="ColorRGB"/> type.
/// </summary>
public static class ColorRGBExtensions
{
    /// <inheritdoc cref="ColorRGB.Add(ColorRGB, ColorRGB)" />
    public static ColorRGB Add(this ColorRGB color, ColorRGB value) => ColorRGB.Add(color, value);

    /// <inheritdoc cref="ColorRGB.Subtract(ColorRGB, ColorRGB)" />
    public static ColorRGB Subtract(this ColorRGB color, ColorRGB value) => ColorRGB.Subtract(color, value);

    /// <inheritdoc cref="ColorRGB.Multiply(ColorRGB, ColorRGB)" />
    public static ColorRGB Multiply(this ColorRGB color, float value) => ColorRGB.Multiply(color, value);

    /// <inheritdoc cref="ColorRGB.Divide(ColorRGB, ColorRGB)" />
    public static ColorRGB Divide(this ColorRGB color, float value) => ColorRGB.Divide(color, value);

    /// <inheritdoc cref="ColorRGB.FromTo(ColorRGB, ColorRGB)" />
    public static ColorRGB FromTo(this ColorRGB from, ColorRGB to) => ColorRGB.FromTo(from, to);

    /// <inheritdoc cref="ColorRGB.Lerp(ColorRGB, ColorRGB, float)" />
    public static ColorRGB Lerp(this ColorRGB from, ColorRGB to, float t) => ColorRGB.Lerp(from, to, t);
}
