namespace Syntriax.Engine.Core;

/// <summary>
/// Represents an RGBA color.
/// </summary>
/// <param name="r">Red value of the <see cref="ColorRGBA"/>.</param>
/// <param name="g">Green value of the <see cref="ColorRGBA"/>.</param>
/// <param name="b">Blue value of the <see cref="ColorRGBA"/>.</param>
/// <param name="a">Alpha value of the <see cref="ColorRGBA"/>.</param>
/// <remarks>
/// Initializes a new instance of the <see cref="ColorRGBA"/> struct with the specified values.
/// </remarks>
[System.Diagnostics.DebuggerDisplay("{ToString(),nq}")]
public readonly struct ColorRGBA(byte r, byte g, byte b, byte a = 255)
{
    /// <summary>
    /// The Red value of the <see cref="ColorRGBA"/>.
    /// </summary>
    public readonly byte R = r;

    /// <summary>
    /// The Green value of the <see cref="ColorRGBA"/>.
    /// </summary>
    public readonly byte G = g;

    /// <summary>
    /// The Blue value of the <see cref="ColorRGBA"/>.
    /// </summary>
    public readonly byte B = b;

    /// <summary>
    /// The Alpha value of the <see cref="ColorRGBA"/>.
    /// </summary>
    public readonly byte A = a;

    public static ColorRGBA operator -(ColorRGBA color) => new((byte)(255 - color.R), (byte)(255 - color.G), (byte)(255 - color.B), color.A);
    public static ColorRGBA operator +(ColorRGBA left, ColorRGBA right) => new((byte)(left.R + right.R).Clamp(0, 255), (byte)(left.G + right.G).Clamp(0, 255), (byte)(left.B + right.B).Clamp(0, 255), (byte)(left.A + right.A).Clamp(0, 255));
    public static ColorRGBA operator -(ColorRGBA left, ColorRGBA right) => new((byte)(left.R - right.R).Clamp(0, 255), (byte)(left.G - right.G).Clamp(0, 255), (byte)(left.B - right.B).Clamp(0, 255), (byte)(left.A - right.A).Clamp(0, 255));
    public static ColorRGBA operator *(ColorRGBA left, ColorRGBA right) => new((byte)(left.R * right.R).Clamp(0, 255), (byte)(left.G * right.G).Clamp(0, 255), (byte)(left.B * right.B).Clamp(0, 255), (byte)(left.A * right.A).Clamp(0, 255));
    public static ColorRGBA operator *(ColorRGBA color, float value) => new((byte)(color.R * value).Clamp(0, 255), (byte)(color.G * value).Clamp(0, 255), (byte)(color.B * value).Clamp(0, 255), (byte)(color.A * value).Clamp(0, 255));
    public static ColorRGBA operator *(float value, ColorRGBA color) => new((byte)(color.R * value).Clamp(0, 255), (byte)(color.G * value).Clamp(0, 255), (byte)(color.B * value).Clamp(0, 255), (byte)(color.A * value).Clamp(0, 255));
    public static ColorRGBA operator /(ColorRGBA color, float value) => new((byte)(color.R / value).Clamp(0, 255), (byte)(color.G / value).Clamp(0, 255), (byte)(color.B / value).Clamp(0, 255), (byte)(color.A / value).Clamp(0, 255));
    public static bool operator ==(ColorRGBA left, ColorRGBA right) => left.R == right.R && left.G == right.G && left.B == right.B && left.A == right.A;
    public static bool operator !=(ColorRGBA left, ColorRGBA right) => left.R != right.R || left.G != right.G || left.B != right.B || left.A != right.A;

    public static implicit operator ColorRGBA(ColorRGB rgb) => new(rgb.R, rgb.G, rgb.B, 255);
    public static implicit operator ColorRGBA(ColorHSV hsv) => (ColorRGB)hsv;

    /// <summary>
    /// Inverts the given <see cref="ColorRGBA"/>.
    /// </summary>
    /// <param name="color">The <see cref="ColorRGBA"/>.</param>
    /// <returns>The inverted <see cref="ColorRGBA"/>.</returns>
    public static ColorRGBA Invert(ColorRGBA color) => -color;

    /// <summary>
    /// Adds two <see cref="ColorRGBA"/>s.
    /// </summary>
    /// <param name="left">The first <see cref="ColorRGBA"/>.</param>
    /// <param name="right">The second <see cref="ColorRGBA"/>.</param>
    /// <returns>The sum of the two <see cref="ColorRGBA"/>s.</returns>
    public static ColorRGBA Add(ColorRGBA left, ColorRGBA right) => left + right;

    /// <summary>
    /// Subtracts one <see cref="ColorRGBA"/> from another.
    /// </summary>
    /// <param name="left">The <see cref="ColorRGBA"/> to subtract from.</param>
    /// <param name="right">The <see cref="ColorRGBA"/> to subtract.</param>
    /// <returns>The result of subtracting the second <see cref="ColorRGBA"/> from the first.</returns>
    public static ColorRGBA Subtract(ColorRGBA left, ColorRGBA right) => left - right;

    /// <summary>
    /// Multiplies a <see cref="ColorRGBA"/> by a scalar value.
    /// </summary>
    /// <param name="color">The <see cref="ColorRGBA"/>.</param>
    /// <param name="value">The scalar value.</param>
    /// <returns>The result of multiplying the <see cref="ColorRGBA"/> by the scalar value.</returns>
    public static ColorRGBA Multiply(ColorRGBA color, float value) => color * value;

    /// <summary>
    /// Divides a <see cref="ColorRGBA"/> by a scalar value.
    /// </summary>
    /// <param name="color">The <see cref="ColorRGBA"/>.</param>
    /// <param name="value">The scalar value.</param>
    /// <returns>The result of dividing the <see cref="ColorRGBA"/> by the scalar value.</returns>
    public static ColorRGBA Divide(ColorRGBA color, float value) => color / value;

    /// <summary>
    /// Calculates the <see cref="ColorRGBA"/> from one point to another.
    /// </summary>
    /// <param name="from">The starting point.</param>
    /// <param name="to">The ending point.</param>
    /// <returns>The <see cref="ColorRGBA"/> from the starting point to the ending point.</returns>
    public static ColorRGBA FromTo(ColorRGBA from, ColorRGBA to) => to - from;

    /// <summary>
    /// Performs linear interpolation between two <see cref="ColorRGBA"/>s.
    /// </summary>
    /// <param name="from">The starting <see cref="ColorRGBA"/> (t = 0).</param>
    /// <param name="to">The ending <see cref="ColorRGBA"/> (t = 1).</param>
    /// <param name="t">The interpolation parameter.</param>
    /// <returns>The interpolated <see cref="ColorRGBA"/>.</returns>
    public static ColorRGBA Lerp(ColorRGBA from, ColorRGBA to, float t) => from + FromTo(from, to) * t;

    /// <summary>
    /// Determines whether the specified object is equal to the current <see cref="ColorRGBA"/>.
    /// </summary>
    /// <param name="obj">The object to compare with the current <see cref="ColorRGBA"/>.</param>
    /// <returns><see cref="true"/> if the specified object is equal to the current <see cref="ColorRGBA"/>; otherwise, <see cref="false"/>.</returns>
    public override bool Equals(object? obj) => obj is ColorRGBA colorRGBA && R.Equals(colorRGBA.R) && G.Equals(colorRGBA.G) && B.Equals(colorRGBA.B) && A.Equals(colorRGBA.A);

    /// <summary>
    /// Generates a hash code for the <see cref="ColorRGBA"/>.
    /// </summary>
    /// <returns>A hash code for the <see cref="ColorRGBA"/>.</returns>
    public override int GetHashCode() => System.HashCode.Combine(R, G, B, A);

    /// <summary>
    /// Converts the <see cref="ColorRGBA"/> to its string representation.
    /// </summary>
    /// <returns>A string representation of the <see cref="ColorRGBA"/>.</returns>
    public override string ToString() => $"{nameof(ColorRGBA)}({R}, {G}, {B}, {A})";
}

/// <summary>
/// Provides extension methods for <see cref="ColorRGBA"/> type.
/// </summary>
public static class ColorRGBAExtensions
{
    /// <inheritdoc cref="ColorRGBA.Add(ColorRGBA, ColorRGBA)" />
    public static ColorRGBA Add(this ColorRGBA color, ColorRGBA value) => ColorRGBA.Add(color, value);

    /// <inheritdoc cref="ColorRGBA.Subtract(ColorRGBA, ColorRGBA)" />
    public static ColorRGBA Subtract(this ColorRGBA color, ColorRGBA value) => ColorRGBA.Subtract(color, value);

    /// <inheritdoc cref="ColorRGBA.Multiply(ColorRGBA, ColorRGBA)" />
    public static ColorRGBA Multiply(this ColorRGBA color, float value) => ColorRGBA.Multiply(color, value);

    /// <inheritdoc cref="ColorRGBA.Divide(ColorRGBA, ColorRGBA)" />
    public static ColorRGBA Divide(this ColorRGBA color, float value) => ColorRGBA.Divide(color, value);

    /// <inheritdoc cref="ColorRGBA.FromTo(ColorRGBA, ColorRGBA)" />
    public static ColorRGBA FromTo(this ColorRGBA from, ColorRGBA to) => ColorRGBA.FromTo(from, to);

    /// <inheritdoc cref="ColorRGBA.Lerp(ColorRGBA, ColorRGBA, float)" />
    public static ColorRGBA Lerp(this ColorRGBA from, ColorRGBA to, float t) => ColorRGBA.Lerp(from, to, t);
}
